/*
 *  Driver Module for Alcatel SpeedTouch USB xDSL modem
 *  Copyright 2001, Alcatel
 *  Written by Johan Verrept (Johan.Verrept@advalvas.be)
 *  Modified by Josep Comas (jcomas@gna.es) to use with 3Com ADSL Modem USB, 19/8/2002
 *  Malloc changes by Alberto Jimnez (ajimenezg@users.sourceforge.net)
 *  OAM support arranged by David Lazaro (dlazaro@well.com)
 */

#include <linux/module.h> 
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/smp_lock.h>

#include <linux/atm.h>
#include <linux/atmdev.h>

#include "sarlib.h"
//#include "version.h"
/*
#define DEBUG 1
#define DEBUG_PACKET 1
*/

#ifdef DEBUG
#define PDEBUG(arg...)  printk(KERN_DEBUG "3Com ADSL Modem USB: " arg)
#else
#define PDEBUG(arg...)
#endif


#ifdef DEBUG_PACKET
#define PACKETDEBUG(arg...) udsl_print_packet ( arg )
#else
#define PACKETDEBUG(arg...)
#endif

#define NORMAL_SPEED  0
#define MAX_SPEED     1

#define MOD_MAJOR_VERSION            0
#define MOD_MINOR_VERSION            4
#define CP4218_VENDORID              0x0506 
#define CP4218_PRODUCTID             0xf003

#define MAX_UDSL                              1
#define UDSL_OBUF_SIZE                    32768
#define UDSL_MINOR                           48
#define UDSL_NUMBER_RCV_URBS                  1
#define UDSL_NUMBER_SND_URBS                  1
#define UDSL_RECEIVE_BUFFER_SIZE          64*53
#define UDSL_MAX_AAL5_MRU                  2048  /* max should be (1500 IP mtu + 2 ppp bytes + 32 * 5 cellheader overhead) for PPPoA and (1500 + 14 + 32*5 cellheader overhead) for PPPoE */
#define UDSL_SEND_CONTEXTS                    8

#define UDSL_POLL_MONITOR_DELAY             100
#define UDSL_POLL_MODEM_STATUS               10
#define UDSL_SEND_BULK_TIMEOUT              100
#define UDSL_READ_BULK_TIMEOUT              500
#define UDSL_WRITE_STATUS_RETRIES             4
#define UDSL_READ_STATUS_RETRIES              4
#define Firmware_Major_Version           0x1cba
#define Firmware_Minor_Version           0x1cbb
#define ADSL_OpenCommand                 0x1cb3
#define DSP_Booted                       0x1cbf
#define LED_REGISTER                     0x7826
#define LED_USB_ON                         0x08
#define LED_ADSL_ON                        0x10
#define LED_DATA_ON                        0x20
#define ON                                    1
#define OFF                                   0
#define I2CS                             0x7fa5
#define I2DAT                            0x7fa6
#define I2CS_START                         0x80
#define I2CS_STOP                          0x40
#define I2CS_LASTREAD                      0x20
#define I2CS_DONE                          0x01
#define I2CS_ACK                           0x02
#define I2CS_WRITE_ADDRESS                 0xa0
#define I2CS_READ_ADDRESS                  0xa1
#define TX_COMMAND                       0x8000
#define RX_COMMAND                       0x8001
#define MODEM_STATUS                     0x8100
#define MODEM_DOWN                         0x00
#define MODEM_WAIT                         0x08
#define MODEM_INIT                         0x10
#define MODEM_UP                           0x20
#define ADSL_Operational_Mode_Ack          0x12
#define ADSL_Connection_Opened             0x24
#define ADSL_Connection_Closed             0x25
#define ADSL_Connection_Lost               0x41
#define ADSL_In_Training                   0x5f
#define ADSL_Open                          0x03
#define ADSL_Close                         0x04
#define ADSL_Open_ANSI                     0x05
#define ADSL_Open_GDMT                     0x06
#define ADSL_Open_GLITE                    0x07
#define ADSL_Open_MULTIMODE                0x08
#define SW_Version_LSB                   0x81fe
#define SW_Version_MSB                   0x81ff
#define R_Operational_Mode               0x80e0
#define R_noiseMarginDnstr               0x810a
#define R_OutputPowerUpstr               0x810b
#define R_attenuationDnStr               0x810c
#define R_noiseMarginUpstr               0x8112
#define R_OutputPowerDnstr               0x8113
#define R_attenuationUpStr               0x8114
#define R_ChanDataIntNear_LSB            0x81a0
#define R_ChanDataIntNear_MSB            0x81a1
#define R_ChanDataFastNear_LSB           0x81a2
#define R_ChanDataFastNear_MSB           0x81a3
#define R_ChanDataIntFar_LSB             0x81a4
#define R_ChanDataIntFar_MSB             0x81a5
#define R_ChanDataFastFar_LSB            0x81a6
#define R_ChanDataFastFar_MSB            0x81a7
#define R_near_end_fec_ni_1              0x81b0
#define R_near_end_fec_ni_2              0x81b1
#define R_near_end_fec_i_1               0x81b2
#define R_near_end_fec_i_2               0x81b3
#define R_near_end_crc_ni_1              0x81b4
#define R_near_end_crc_ni_2              0x81b5
#define R_near_end_crc_i_1               0x81b6
#define R_near_end_crc_i_2               0x81b7
#define R_near_end_hec_ni_1              0x81b8
#define R_near_end_hec_ni_2              0x81b9
#define R_near_end_hec_i_1               0x81ba
#define R_near_end_hec_i_2               0x81bb
#define R_far_end_fec_ni_1               0x81bc
#define R_far_end_fec_ni_2               0x81bd
#define R_far_end_fec_i_1                0x81be
#define R_far_end_fec_i_2                0x81bf
#define R_far_end_crc_ni_1               0x81c0
#define R_far_end_crc_ni_2               0x81c1
#define R_far_end_crc_i_1                0x81c2
#define R_far_end_crc_i_2                0x81c3
#define R_far_end_hec_ni_1               0x81c4
#define R_far_end_hec_ni_2               0x81c5
#define R_far_end_hec_i_1                0x81c6
#define R_far_end_hec_i_2                0x81c7

#define FastPath                              1
#define InterleavedPath                       2

#define ATM_PT_TYPE                         0x4
#define OAM_LOOPBACK                       0x18  /* OAM type = 0001 (Fault Management) + Function type = 1000 (OAM Cell Loopback) */

/* endpoint declarations */

#define UDSL_ENDPOINT_INFO_OUT      0x1
#define UDSL_ENDPOINT_INFO_IN      0x81
#define UDSL_ENDPOINT_DATA_OUT      0x2
#define UDSL_ENDPOINT_DATA_IN      0x82


static int speed = NORMAL_SPEED;
static int open = ADSL_Open_ANSI;

MODULE_AUTHOR ("Johan Verrept (Johan.Verrept@advalvas.be) for Alcatel SpeedTouch USB, Josep Comas (jcomas@gna.es) for 3Com ADSL Modem USB (3CP4218)");
MODULE_DESCRIPTION ("Driver for 3Com ADSL Modem USB (3CP4218)");
MODULE_LICENSE ("GPL");
MODULE_PARM(speed, "i");
MODULE_PARM_DESC(speed, "Driver speed: 0 = normal, 1 = fast (default = 0)");
MODULE_PARM(open, "i");
MODULE_PARM_DESC(open, "Open mode: 5 = ANSI T1.413, 7 = ITU-T G.992.2 (G.LITE), 6 = ITU-T G.992.1 (G.DMT), 8 = multimode, 3 = other (default = 5)");


/* usb_device_id struct */

static struct usb_device_id udsl_usb_ids [] = 
{
  { USB_DEVICE (CP4218_VENDORID, CP4218_PRODUCTID) },
  { } /* list terminator */
};

/* not exporting this prevents the depmod from generating the map that causes the modules to be isnserted as driver.
 * we do not want this, we want the script run.
MODULE_DEVICE_TABLE ( usb, udsl_usb_ids);
*/
/* context declarations */

typedef struct udsl_data_ctx {
  struct sk_buff            *skb;
  urb_t                     *urb;
  struct udsl_instance_data *instance;
  unsigned char *pdata;
  } udsl_data_ctx_t;

typedef struct udsl_usb_send_data_context {
  urb_t          urb;
  struct sk_buff *skb;
  struct atm_vcc *vcc;
  struct udsl_instance_data *instance;  
  unsigned char *pdata;
  } udsl_usb_send_data_context_t;

/*
 * UDSL main driver data
 */ 

struct udsl_modem_info {
  unsigned char lastcmd;
  unsigned char laststatus;
  unsigned char modminor;
  unsigned char modmajor;
  unsigned char modspeed;
  unsigned char firmminor;
  unsigned char firmmajor;
  unsigned char hardversion;
  unsigned char harddata[4];
  unsigned char mac[6];
  unsigned char serial[12];
  char alcfirmminor;
  char alcfirmmajor;
  short vpi;
  int vci;
  unsigned int atmcelltx;
  unsigned int atmcellrx;
  unsigned int oamloopback;
  unsigned int downbitrate;
  unsigned int upbitrate;
  unsigned char openmode;
  unsigned char operationalmode;
  unsigned char linestatus;
  unsigned char datapath;
  unsigned int linklost;
  unsigned char extendedinfo;
  unsigned short downcorrectederrors;
  unsigned short downuncorrectederrors;
  unsigned short downhecerrors;
  unsigned short upcorrectederrors;
  unsigned short upuncorrectederrors;
  unsigned short uphecerrors;
  unsigned char downattenuation;
  unsigned char upattenuation;
  signed char downsignalnoiseratio;
  signed char upsignalnoiseratio;
  signed char downoutputpower;
  signed char upoutputpower;
  } udsl_modem_info_t;

#define UDSL_IOCTL_START                      1  /* start transfering ATM cells */
#define UDSL_IOCTL_STOP                       2  /* stop transfering ATM cells */
#define UDSL_IOCTL_INFO  _IOWR('U', 3, struct udsl_modem_info)  /* get info about modem and ADSL */
#define UDSL_IOCTL_RECONNECT                  4  /* use to reconnect ADSL line when you think is down in max speed mode */
#define UDSL_IOCTL_UPDATE_OPS_MODE            5  /* use when you think that ADSL line is up in max speed mode */
#define UDSL_IOCTL_START_EXTENDED_INFO        6  /* start getting additional info if driver is in normal speed mode */
#define UDSL_IOCTL_STOP_EXTENDED_INFO         7  /* stop get additional info */
#define UDSL_IOCTL_CLEAR_COUNTERS             8  /* clear counters (cells trasmited, received, ...) */

struct udsl_instance_data {
  int                 minor;

  /* usb device part */
  struct usb_device  *usb_dev;
  udsl_data_ctx_t    *rcvbufs;
  struct sk_buff_head sndqueue;
  spinlock_t          sndqlock;
  udsl_usb_send_data_context_t send_ctx[UDSL_NUMBER_SND_URBS];
  int                 data_started;

  unsigned char oambuff[64];
  struct udsl_modem_info modem_info;
  int dataled;
  int adslled;
  int pending_data_led;
 
  /* atm device part */
  struct atm_dev     *atm_dev;
  struct sk_buff_head recvqueue;
  spinlock_t          recvqlock;
    
  struct sarlib_vcc_data *sarlib_vcc_list;
  } udsl_instance_t;

struct udsl_instance_data *minor_data[MAX_UDSL];

static const char udsl_driver_name[] = "3Com ADSL Modem USB";

/* data thread */
static int datapid = 0;
static int datapid2 = 0;
static int datapid3 = 0;
DECLARE_WAIT_QUEUE_HEAD(udsl_wqh);
DECLARE_WAIT_QUEUE_HEAD(udsl_wqh2);
DECLARE_WAIT_QUEUE_HEAD(udsl_wqh3);
DECLARE_WAIT_QUEUE_HEAD(udsl_wqh4);

#ifdef DEBUG_PACKET
int udsl_print_packet (const unsigned char *data, int len);
#endif

int udsl_usb_close_line (struct udsl_instance_data *instance);
unsigned char *pad_data(unsigned char *source, int l); 
int remove_control_cells(unsigned char *source, int l, unsigned char *out, struct udsl_instance_data *instance);

/*
 * atm driver prototypes and stuctures
 */

static int udsl_atm_open(struct atm_vcc *vcc, short vpi, int vci);
static void udsl_atm_close(struct atm_vcc *vcc);
static int udsl_atm_ioctl(struct atm_dev *dev, unsigned int cmd, void *arg);
static int udsl_atm_send(struct atm_vcc *vcc, struct sk_buff *skb);
int udsl_atm_proc_read (struct atm_dev * atm_dev, loff_t * pos, char * page);
void udsl_atm_processqueue ( struct udsl_instance_data *instance);

static struct atmdev_ops udsl_atm_devops = 
  {
  open:       udsl_atm_open,
  close:      udsl_atm_close,
  ioctl:      udsl_atm_ioctl,
  send:       udsl_atm_send,
  proc_read:  udsl_atm_proc_read,
  };

typedef struct udsl_atm_dev_data {
  struct sarlib_vcc_data *sarlib_vcc;
  } udsl_atm_dev_data_t;

/*
 * usb driver prototypes and structures
 */
static void * udsl_usb_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id);
static void   udsl_usb_disconnect(struct usb_device *dev, void *ptr);
       int    udsl_usb_send_data (struct udsl_instance_data *instance, struct atm_vcc *vcc, struct sk_buff *skb);
static int    udsl_usb_ioctl (struct usb_device *hub, unsigned int code, void *user_data);
static int    udsl_usb_cancelsends (struct udsl_instance_data *instance, struct atm_vcc *vcc);

static struct usb_driver udsl_usb_driver =
{
    name:       udsl_driver_name,
    probe:      udsl_usb_probe,
    disconnect: udsl_usb_disconnect,
    ioctl:      udsl_usb_ioctl,
    id_table:   udsl_usb_ids,
};

/************
**   ATM   **
************/

/***************************************************************************
*
* init functions
*
****************************************************************************/

struct atm_dev *udsl_atm_startdevice(struct udsl_instance_data *instance, struct atmdev_ops *devops) {
  MOD_INC_USE_COUNT;

  instance->atm_dev = atm_dev_register(udsl_driver_name, devops, -1, 0);
  instance->atm_dev->dev_data = instance;
  instance->atm_dev->ci_range.vpi_bits = ATM_CI_MAX;
  instance->atm_dev->ci_range.vci_bits = ATM_CI_MAX;
  instance->atm_dev->signal = ATM_PHY_SIG_LOST;
  
  skb_queue_head_init(&instance->recvqueue);

  /* tmp init atm device, set to 128kbit */
  instance->atm_dev->link_rate = 128*1000/424;
  
  return instance->atm_dev;
  }

void udsl_atm_stopdevice(struct udsl_instance_data *instance) {
  struct atm_vcc *walk;
  struct sk_buff *skb;
  struct atm_dev *atm_dev;
  unsigned long iflags;
    
  if (!instance->atm_dev)
    return;

  atm_dev = instance->atm_dev;

  /* clean queue */
  spin_lock_irqsave(&instance->recvqlock, iflags);  
  while (!skb_queue_empty(&instance->recvqueue)) {
    skb = skb_dequeue (&instance->recvqueue);
    dev_kfree_skb(skb);
    };
  spin_unlock_irqrestore (&instance->recvqlock, iflags);
  
  atm_dev->signal = ATM_PHY_SIG_LOST;
  walk = atm_dev->vccs;
  shutdown_atm_dev(atm_dev);

  for (; walk; walk = walk->next)
    wake_up(&walk->sleep);

  instance->atm_dev = NULL;  
  MOD_DEC_USE_COUNT;
  }

void udsl_atm_set_mac(struct udsl_instance_data *instance, const char mac[6]) {
  if (!instance->atm_dev)
      return;

  memcpy (instance->atm_dev->esi, mac, 6);
  }

/***************************************************************************
*
* ATM helper functions
*
****************************************************************************/
struct sk_buff * udsl_atm_alloc_tx(struct atm_vcc *vcc,unsigned int size) {
  sarlib_vcc_data_t *sarlib_vcc = ((struct udsl_atm_dev_data *)vcc->dev_data)->sarlib_vcc;
  
  if (sarlib_vcc)
    return sarlib_alloc_tx(sarlib_vcc, size);
  
  printk(KERN_INFO "3Com ADSL Modem USB: udsl_atm_alloc_tx could not find correct alloc_tx function !\n");
  return NULL;
  }

int udsl_atm_proc_read (struct atm_dev * atm_dev, loff_t * pos, char * page) {
  struct udsl_instance_data *instance = (struct udsl_instance_data *) atm_dev->dev_data;
  int left = *pos;
    
  if (!left--)
    return sprintf(page, "3Com ADSL Modem USB:%d (%02x:%02x:%02x:%02x:%02x:%02x)\n", instance->minor,
              atm_dev->esi[0], atm_dev->esi[1], atm_dev->esi[2], atm_dev->esi[3], atm_dev->esi[4], atm_dev->esi[5]);
  
  if (!left--)
    return sprintf(page, "AAL0: tx %d ( %d err ), rx %d ( %d err, %d drop )\n",
      atomic_read(&atm_dev->stats.aal0.tx), atomic_read(&atm_dev->stats.aal0.tx_err),
      atomic_read(&atm_dev->stats.aal0.rx), atomic_read(&atm_dev->stats.aal0.rx_err), atomic_read(&atm_dev->stats.aal0.rx_drop));

  if (!left--)
    return sprintf(page, "AAL5: tx %d ( %d err ), rx %d ( %d err, %d drop )\n",
      atomic_read(&atm_dev->stats.aal5.tx), atomic_read(&atm_dev->stats.aal5.tx_err),
      atomic_read(&atm_dev->stats.aal5.rx), atomic_read(&atm_dev->stats.aal5.rx_err), atomic_read(&atm_dev->stats.aal5.rx_drop));

  return 0;
  }

/***************************************************************************
*
* ATM DATA functions
*
****************************************************************************/
int udsl_atm_send(struct atm_vcc *vcc, struct sk_buff *skb) {
  struct udsl_atm_dev_data  *dev_data = (struct udsl_atm_dev_data *)  vcc->dev_data;
  struct udsl_instance_data *instance = (struct udsl_instance_data *) vcc->dev->dev_data;
  struct sk_buff *new = NULL;
  int err;
  
  PDEBUG("udsl_atm_send called\n");
  
  if (!dev_data)
    return -EINVAL;
  
  switch (vcc->qos.aal) {
    case ATM_AAL5:
      new = sarlib_encode_aal5(dev_data->sarlib_vcc, skb);
      if (!new) goto nomem;
      if (new != skb) {
        vcc->pop(vcc, skb);
        skb = new;
        }
      new = sarlib_encode_rawcell(dev_data->sarlib_vcc, skb);
      if (!new) goto nomem;
      if (new != skb) {
        vcc->pop(vcc, skb);
        skb = new;
        }
      err = udsl_usb_send_data (instance, vcc, skb);
      PDEBUG("udsl_atm_send successfull (%d)\n", err);
      return err;
      break;
    default:
      return -EINVAL;
    };      
    
  PDEBUG("udsl_atm_send unsuccessfull\n");    
  return 0;
nomem:
  vcc->pop(vcc, skb);
  return -ENOMEM;
};


void udsl_atm_processqueue ( struct udsl_instance_data *instance) {
  struct sarlib_vcc_data *sarlib_vcc = NULL;
  struct sk_buff *new = NULL, *skb = NULL, *tmp = NULL;
  unsigned long iflags;
  
  /* quick check */
  spin_lock_irqsave (&instance->recvqlock, iflags);
  if (skb_queue_empty(&instance->recvqueue)) {
    spin_unlock_irqrestore (&instance->recvqlock, iflags);
    return;
    }
  PDEBUG("udsl_atm_processqueue entered\n");

  while (!skb_queue_empty(&instance->recvqueue)) {
    skb = skb_dequeue (&instance->recvqueue);
    
    spin_unlock_irqrestore (&instance->recvqlock, iflags);
    
    PDEBUG("skb = %p, skb->len = %d\n", skb, skb->len);

    PACKETDEBUG (skb->data, skb->len);

    while ( (new = sarlib_decode_rawcell (instance->sarlib_vcc_list, skb, &sarlib_vcc)) != NULL) {
      PDEBUG("(after cell processing)skb->len = %d\n", new->len);
      switch (sarlib_vcc->type) {
        case SARLIB_TYPE_AAL5:
          tmp = new;        
          new = sarlib_decode_aal5 (sarlib_vcc, new);

          /* we can't send NULL skbs upstream, the ATM layer would try to close the vcc... */
          if (new) {
            PDEBUG("(after aal5 decap) skb->len = %d\n", new->len);
            if (new->len && atm_charge(sarlib_vcc->vcc, new->truesize)) {
              PACKETDEBUG (new->data, new->len);
              sarlib_vcc->vcc->push(sarlib_vcc->vcc, new);
              }
            else {
              PDEBUG("dropping incoming packet : rx_inuse = %d, vcc->sk->rcvbuf = %d, skb->true_size = %d\n", atomic_read(&sarlib_vcc->vcc->rx_inuse), sarlib_vcc->vcc->sk->rcvbuf, new->truesize);
              dev_kfree_skb(new);
              }
            }
          else {
            PDEBUG("sarlib_decode_aal5 returned NULL!\n");
            dev_kfree_skb(tmp);
          }
          break;
        default:
          /* not supported. we delete the skb. */
          printk(KERN_INFO "3Com ADSL Modem USB: illegal vcc type. Dropping packet.\n");
          dev_kfree_skb(new);
          break;
        }
      };
    dev_kfree_skb (skb);
    spin_lock_irqsave (&instance->recvqlock, iflags);
    };

  spin_unlock_irqrestore (&instance->recvqlock, iflags);
  PDEBUG("udsl_atm_processqueue successfull\n");
  }

int udsl_atm_processqueue_thread (void *data) {
  int i=0;
  DECLARE_WAITQUEUE(wait, current);
     
  lock_kernel();
  exit_files(current);  /* daemonize doesn't do exit_files */
  daemonize();

  /* Setup a nice name */
  strcpy(current->comm, "kSpeedSARd");
  
  add_wait_queue(&udsl_wqh, &wait);

  for(;;) {
    interruptible_sleep_on(&udsl_wqh);
    if (signal_pending(current)) break;
    PDEBUG("SpeedSARd awoke\n");
    for (i=0; i < MAX_UDSL ; i++)
      if (minor_data[i])
        udsl_atm_processqueue(minor_data[i]);
    };
  
  datapid = 0;
  PDEBUG("SpeedSARd is exiting\n");
  return 0;
  }


void udsl_atm_sar_start (void) {
  datapid = kernel_thread(udsl_atm_processqueue_thread, (void *)NULL,
      CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
  }

void udsl_atm_sar_stop (void) {
  int ret;

  /* Kill the thread */
  ret = kill_proc(datapid, SIGTERM, 1);
  if (!ret) {
    /* Wait 10 seconds */
    int count = 10 * 100;

    while (datapid && --count) {
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(1);
      }

    if (!count)
      err("giving up on killing SpeedSAR thread.");
    }
  }

/***************************************************************************
*
* SAR driver entries
*
****************************************************************************/
int udsl_atm_open(struct atm_vcc *vcc, short vpi, int vci) {
  struct udsl_atm_dev_data  *dev_data;
  struct udsl_instance_data *instance = (struct udsl_instance_data *)vcc->dev->dev_data;

  PDEBUG("udsl_atm_open called\n");

  /* at the moment only AAL5 support */
  if (vcc->qos.aal != ATM_AAL5)
    return -EINVAL;

  MOD_INC_USE_COUNT;
  dev_data = (struct udsl_atm_dev_data *)kmalloc (sizeof(struct udsl_atm_dev_data), GFP_ATOMIC);
  if (!dev_data)
    return -ENOMEM;

  dev_data->sarlib_vcc = sarlib_open (&(instance->sarlib_vcc_list), vcc, SARLIB_TYPE_AAL5, vpi, vci, 0, 0, SARLIB_USE_53BYTE_CELL | SARLIB_SET_PTI);
  if (!dev_data->sarlib_vcc) {
    kfree(dev_data);
    return -ENOMEM; /* this is the only reason sarlib_open can fail... */
    }

  vcc->vpi = vpi;
  vcc->vci = vci;
  instance->modem_info.vpi = vpi;
  instance->modem_info.vci = vci;
  set_bit(ATM_VF_ADDR,    &vcc->flags);
  set_bit(ATM_VF_PARTIAL, &vcc->flags);
  set_bit(ATM_VF_READY,   &vcc->flags);
  vcc->dev_data = dev_data;
  vcc->alloc_tx = udsl_atm_alloc_tx;
  
  dev_data->sarlib_vcc->mtu = UDSL_MAX_AAL5_MRU;
  
  PDEBUG("udsl_atm_open successfull\n");
  return 0;
}

void udsl_atm_close(struct atm_vcc *vcc) {
  struct udsl_atm_dev_data  *dev_data = (struct udsl_atm_dev_data *)vcc->dev_data;
  struct udsl_instance_data *instance = (struct udsl_instance_data *)vcc->dev->dev_data;
  
  PDEBUG("udsl_atm_close called\n");

  /* freeing resources */
  /* cancel all sends on this vcc */
  udsl_usb_cancelsends(instance, vcc);

  sarlib_close (&(instance->sarlib_vcc_list), dev_data->sarlib_vcc);
  kfree (dev_data);
  vcc->dev_data = NULL;
  clear_bit(ATM_VF_PARTIAL, &vcc->flags);

  /* freeing address */
  vcc->vpi = ATM_VPI_UNSPEC;
  vcc->vci = ATM_VCI_UNSPEC;
  instance->modem_info.vpi = 0;
  instance->modem_info.vci = 0;
  clear_bit(ATM_VF_ADDR,    &vcc->flags);

  MOD_DEC_USE_COUNT;
  
  PDEBUG("udsl_atm_close successfull\n");  
  return;
};

int udsl_atm_ioctl(struct atm_dev *dev, unsigned int cmd, void *arg) {
  switch (cmd) {
    case ATM_QUERYLOOP:
      return put_user(ATM_LM_NONE, (int *) arg) ? -EFAULT : 0;
    default:
      return -ENOIOCTLCMD;
    }
};


/************
**   USB   **
************/

/***************************************************************************
*
* usb data functions
*
****************************************************************************/

typedef struct udsl_cb {
  struct atm_vcc *vcc;
  } udsl_cb_t;

static void udsl_usb_send_data_complete(urb_t *urb) {
  udsl_usb_send_data_context_t *ctx      = (udsl_usb_send_data_context_t *)urb->context;
  struct udsl_instance_data    *instance = ctx->instance;
  int err;
  unsigned long flags;
  
  PDEBUG("udsl_usb_send_data_completion (vcc = %p, skb = %p, status %d)\n", ctx->vcc, ctx->skb, urb->status);
  if (speed != MAX_SPEED && instance->pending_data_led == 0)
    instance->pending_data_led++;
  instance->modem_info.atmcelltx = instance->modem_info.atmcelltx + (urb->actual_length / 64);

  ctx->vcc->pop (ctx->vcc, ctx->skb);
  ctx->skb = NULL;
  kfree(ctx->pdata);

  spin_lock_irqsave (&instance->sndqlock, flags);
  if (skb_queue_empty(&instance->sndqueue)) {
    spin_unlock_irqrestore (&instance->sndqlock, flags);  
    return;
    }

  /* submit next skb */
  ctx->skb = skb_dequeue (&(instance->sndqueue));
  ctx->vcc = ((udsl_cb_t *)(ctx->skb->cb))->vcc;
  ctx->pdata = pad_data(ctx->skb->data, ctx->skb->len);
  spin_unlock_irqrestore (&instance->sndqlock, flags);

  if (!ctx->pdata)
    return;
  
  PACKETDEBUG (ctx->pdata, ((ctx->skb->len)/53)*64);

  FILL_BULK_URB(urb,
    instance->usb_dev,
    usb_sndbulkpipe(instance->usb_dev, UDSL_ENDPOINT_DATA_OUT),
    ctx->pdata,
    ((ctx->skb->len)/53)*64,
    (usb_complete_t)udsl_usb_send_data_complete,
    ctx
    );
  urb->transfer_flags |= USB_QUEUE_BULK;
  
  err = usb_submit_urb(urb);

  PDEBUG("udsl_usb_send_data_completion (send packet %p with length %d), retval = %d\n", ctx->skb, ctx->skb->len, err);
  }

int udsl_usb_cancelsends (struct udsl_instance_data *instance, struct atm_vcc *vcc) {
  int i;
  unsigned long flags;
  
  spin_lock_irqsave (&instance->sndqlock, flags);
  for (i=0; i < UDSL_NUMBER_SND_URBS ; i++) {
    if (!instance->send_ctx[i].skb) continue;
    if (instance->send_ctx[i].vcc==vcc) {
      usb_unlink_urb (&instance->send_ctx[i].urb);
      instance->send_ctx[i].vcc->pop (instance->send_ctx[i].vcc, instance->send_ctx[i].skb);
      instance->send_ctx[i].skb = NULL;
      kfree(instance->send_ctx[i].pdata);
      }
    }
  spin_unlock_irqrestore(&instance->sndqlock, flags);
  
  return 0;
  }

/**** send ******/
int udsl_usb_send_data (struct udsl_instance_data *instance, struct atm_vcc *vcc, struct sk_buff *skb) {
  int    err, i;
  urb_t *urb;
  unsigned long flags;
  
  PDEBUG("udsl_usb_send_data entered, sending packet %p with length %d\n", skb, skb->len);

  if (!instance->data_started)
    return -EAGAIN;

  //PACKETDEBUG (skb->data, skb->len);

  spin_lock_irqsave (&instance->sndqlock, flags);
  ((udsl_cb_t *)skb->cb)->vcc = vcc;
  
  /* we are already queueing */
  if (!skb_queue_empty(&instance->sndqueue)) {
    skb_queue_tail (&instance->sndqueue, skb);
    spin_unlock_irqrestore(&instance->sndqlock, flags);
    PDEBUG("udsl_usb_send_data: already queing, skb (0x%p) queued\n", skb);
    return 0;
    }

  for (i=0; i < UDSL_NUMBER_SND_URBS ; i++)
    if (instance->send_ctx[i].skb==NULL) break;

  /* we must start queueing */
  if (i == UDSL_NUMBER_SND_URBS) {
    skb_queue_tail (&instance->sndqueue, skb);
    spin_unlock_irqrestore(&instance->sndqlock, flags);
    PDEBUG("udsl_usb_send_data: skb (0x%p) queued\n", skb);
    return 0;
    };

  /* init context */
  urb = &(instance->send_ctx[i].urb);
  instance->send_ctx[i].skb       = skb;
  instance->send_ctx[i].vcc       = vcc;
  instance->send_ctx[i].instance  = instance;
  instance->send_ctx[i].pdata = pad_data(skb->data, skb->len);

  spin_unlock_irqrestore(&instance->sndqlock, flags);

  /* submit packet */

  if (!instance->send_ctx[i].pdata)
    return -EAGAIN;

  //PDEBUG("PAD Packet:\n");

  PACKETDEBUG (instance->send_ctx[i].pdata, ((skb->len)/53)*64);

  FILL_BULK_URB(urb,
    instance->usb_dev,
    usb_sndbulkpipe(instance->usb_dev, UDSL_ENDPOINT_DATA_OUT),
    instance->send_ctx[i].pdata,
    ((skb->len)/53)*64,
    (usb_complete_t)udsl_usb_send_data_complete,
    &(instance->send_ctx[i])
    );
  urb->transfer_flags |= USB_QUEUE_BULK;
  
  err = usb_submit_urb(urb);

  if (err != 0)
    skb_unlink(skb);

  PDEBUG("udsl_usb_send_data done (retval = %d)\n", err);
  return err;
  }

/********* receive *******/
void udsl_usb_data_receive(urb_t *urb) {
  udsl_data_ctx_t *ctx; 
  struct udsl_instance_data *instance;
  unsigned long flags;
  int lpdata;
  //DECLARE_WAITQUEUE(wait, current);
  
  if (!urb) {
    wake_up(&udsl_wqh4);
    return;
    }

if (!(urb->status == 0 && urb->actual_length == 0))   
  PDEBUG("udsl_usb_receive_data entered, got packet %p with length %d an status %d\n", urb, urb->actual_length, urb->status);

  ctx = (udsl_data_ctx_t *)urb->context;
  if (!ctx || !ctx->skb || !ctx->pdata) {
    wake_up(&udsl_wqh4);
    return;
    }

  instance = ctx->instance;

if (!(urb->status == 0 && urb->actual_length == 0)) {
  switch (urb->status) {
    case 0:
      PDEBUG("udsl_usb_data_receive: processing urb with ctx %p, urb %p (%p), skb %p\n", ctx, ctx ? ctx->urb : NULL , urb, ctx ? ctx->skb : NULL);
      if (speed != MAX_SPEED && instance->pending_data_led == 0)
        instance->pending_data_led++;
      instance->modem_info.atmcellrx = instance->modem_info.atmcellrx + (urb->actual_length / 53);
   
      lpdata = remove_control_cells(ctx->pdata, urb->actual_length, ctx->skb->data, instance);

      if (lpdata > 0) {
        /* update the skb structure */
        //skb_put(ctx->skb, urb->actual_length);
        skb_put(ctx->skb, lpdata);

        /* queue the skb for processing and wake the SAR */
        spin_lock_irqsave (&instance->recvqlock, flags);
        skb_queue_tail (&instance->recvqueue, ctx->skb);
        spin_unlock_irqrestore (&instance->recvqlock, flags);
        wake_up(&udsl_wqh);

        /* get a new skb */
        ctx->skb = dev_alloc_skb(UDSL_RECEIVE_BUFFER_SIZE);
        if (!ctx->skb) {
          PDEBUG("No skb, loosing urb.\n");
          usb_free_urb(ctx->urb);
          ctx->urb = NULL;  
          wake_up(&udsl_wqh4);
          return;
          }
        }
      break;
    case -EPIPE:  /* stall or babble */
      usb_clear_halt(urb->dev, usb_rcvbulkpipe(urb->dev, UDSL_ENDPOINT_DATA_IN));
      break;
    case -ENOENT:  /* buffer was unlinked */
    case -EILSEQ:    /* unplug or timeout */
    case -ETIMEDOUT: /* unplug or timeout */
      /* 
       * we don't do anything here and we don't resubmit
       */
      wake_up(&udsl_wqh4);
      return;
    }
}
else
  {
	  /*
    add_wait_queue(&udsl_wqh4, &wait);
    interruptible_sleep_on_timeout(&udsl_wqh4, UDSL_SEND_BULK_TIMEOUT * HZ / 1000); 
    remove_wait_queue(&udsl_wqh4, &wait);
    */
  }

  FILL_BULK_URB(urb,
    instance->usb_dev,
    usb_rcvbulkpipe(instance->usb_dev, UDSL_ENDPOINT_DATA_IN),
    //(unsigned char *)ctx->skb->data,
    ctx->pdata,
    UDSL_RECEIVE_BUFFER_SIZE,
    (usb_complete_t)udsl_usb_data_receive,
    ctx
    );
  urb->transfer_flags        |= USB_QUEUE_BULK;

  //usb_submit_urb(urb);
  wake_up(&udsl_wqh4);

  return;
  };

int udsl_usb_receive_thread (void *data) {
  urb_t *urb = (urb_t *)data;
  DECLARE_WAITQUEUE(wait, current);
     
  lock_kernel();
  exit_files(current);  /* daemonize doesn't do exit_files */
  daemonize();

  /* Setup a nice name */
  strcpy(current->comm, "kRcvADSLd");

  add_wait_queue(&udsl_wqh4, &wait);
  for(;;) {
    usb_submit_urb(urb);
    interruptible_sleep_on(&udsl_wqh4);
    if (signal_pending(current)) break;
    if (urb->status == 0 && urb->actual_length == 0) {
      interruptible_sleep_on_timeout(&udsl_wqh4, 5 * HZ / 1000);
      if (signal_pending(current)) break;
      }
    }
  remove_wait_queue(&udsl_wqh4, &wait);
  
  datapid3 = 0;
  PDEBUG("kRcvADSLd is exiting\n");
  return 0;
  }

void udsl_usb_rcv_start (void *data) {
  datapid3 = kernel_thread(udsl_usb_receive_thread, data,
      CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
  }

void udsl_usb_rcv_stop (void) {
  int ret;

  /* Kill the thread */
  ret = kill_proc(datapid3, SIGTERM, 1);
  if (!ret) {
    /* Wait 10 seconds */
    int count = 10 * 100;

    while (datapid3 && --count) {
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(1);
      }

    if (!count)
      err("giving up on killing RcvADSL thread.");
    }
  }

int udsl_usb_data_init(struct udsl_instance_data *instance) {
  int i, succes;

  PDEBUG("max packet size on endpoint %d is %d\n", UDSL_ENDPOINT_DATA_OUT, usb_maxpacket (instance->usb_dev, usb_sndbulkpipe(instance->usb_dev, UDSL_ENDPOINT_DATA_OUT), 0));

  instance->rcvbufs = (udsl_data_ctx_t *)kmalloc (sizeof(udsl_data_ctx_t) * UDSL_NUMBER_RCV_URBS , GFP_ATOMIC);
  if (!instance->rcvbufs) 
    return -ENOMEM;  

  memset (instance->rcvbufs, 0, sizeof(udsl_data_ctx_t) * UDSL_NUMBER_RCV_URBS);
  skb_queue_head_init(&instance->sndqueue);

  for (i=0, succes = 0 ; i < UDSL_NUMBER_RCV_URBS; i++) {
    udsl_data_ctx_t *ctx = &(instance->rcvbufs[i]);
    
    ctx->urb = NULL;
    ctx->skb = NULL;
    ctx->pdata = kmalloc(UDSL_RECEIVE_BUFFER_SIZE, GFP_ATOMIC);
    if (!ctx->pdata) continue;
    ctx->skb = dev_alloc_skb(UDSL_RECEIVE_BUFFER_SIZE);
    if (!ctx->skb) {
      kfree(ctx->pdata);
      ctx->pdata = NULL;
      continue;
      }

    ctx->urb = usb_alloc_urb(0);
    if (!ctx->urb) {
      kfree(ctx->pdata);
      kfree_skb(ctx->skb);
      ctx->skb = NULL;
      break;
      };

    FILL_BULK_URB(ctx->urb,
      instance->usb_dev,
      usb_rcvbulkpipe(instance->usb_dev, UDSL_ENDPOINT_DATA_IN),
      //(unsigned char *)ctx->skb->data,
      ctx->pdata,
      UDSL_RECEIVE_BUFFER_SIZE,
      (usb_complete_t)udsl_usb_data_receive,
      ctx
      );
    ctx->urb->transfer_flags |= USB_QUEUE_BULK;

    ctx->instance = instance;
    
    PDEBUG("udsl_usb_data_init: usb with skb->truesize = %d (Asked for %d)\n", ctx->skb->truesize, UDSL_RECEIVE_BUFFER_SIZE);
    udsl_usb_rcv_start(ctx->urb);
   /* 
    if (usb_submit_urb(ctx->urb) < 0)      
      PDEBUG("udsl_usb_data_init: Submit failed, loosing urb.\n");
    else*/
      succes++;
    }
  
  instance->data_started=1;
  instance->atm_dev->signal = ATM_PHY_SIG_FOUND;
  
  PDEBUG("udsl_usb_data_init %d urb%s queued for receive\n", succes, (succes!=1) ? "s" : "");

  //udsl_usb_rcv_start();

  return 0;
  }

int udsl_usb_data_exit(struct udsl_instance_data *instance) {
  int i;
  
  if (!instance->data_started)
    return 0;
  
  if (!instance->rcvbufs)
    return 0;

  udsl_usb_rcv_stop();
   
  /* destroy urbs */
  for (i=0; i < UDSL_NUMBER_RCV_URBS; i++) {
    udsl_data_ctx_t *ctx = &(instance->rcvbufs[i]);

    if ((!ctx->pdata) || (!ctx->urb) || (!ctx->skb))
      continue;
     
    if (ctx->urb->status == -EINPROGRESS)
      usb_unlink_urb(ctx->urb);
      
    usb_free_urb(ctx->urb);
    kfree_skb(ctx->skb);
    ctx->skb = NULL;
    kfree(ctx->pdata);
    ctx->pdata = NULL;
    };
  
  for (i=0; i < UDSL_NUMBER_SND_URBS; i++) {
    udsl_usb_send_data_context_t *ctx = &(instance->send_ctx[i]);

    if (!ctx->skb)
      continue;

    if (ctx->urb.status == -EINPROGRESS)
      usb_unlink_urb(&ctx->urb);
    
    /* not sure if completion handler is called */
    if (ctx->skb)
      ctx->vcc->pop (ctx->vcc, ctx->skb);
    ctx->skb = NULL;
    kfree(&ctx->pdata);
    }
  
  /* free receive contexts */
  kfree (instance->rcvbufs);
  instance->rcvbufs = NULL;
  
  instance->data_started=0;
  instance->atm_dev->signal = ATM_PHY_SIG_LOST;
  
  return 0;
  };

void udsl_usb_send_bulk_complete(urb_t *urb) {

  //PDEBUG("udsl_usb_send_bulk_complete called\n");

  wake_up(&udsl_wqh3);
  }

int udsl_usb_send_bulk(struct udsl_instance_data *instance, char *buf, int len) {
  urb_t *urb;
  int err;
  DECLARE_WAITQUEUE(wait, current);

  //PDEBUG("udsl_usb_send_bulk called\n");

  urb = usb_alloc_urb(0);
  if (!urb)
    return -ENOMEM;

  add_wait_queue(&udsl_wqh3, &wait);

  FILL_BULK_URB(urb,
   instance->usb_dev,
   usb_sndbulkpipe(instance->usb_dev, UDSL_ENDPOINT_INFO_OUT),
   buf,
   len,
   (usb_complete_t)udsl_usb_send_bulk_complete,
   instance
   );

  urb->transfer_flags |= USB_QUEUE_BULK;
  
  err = usb_submit_urb(urb);
  if (err) {
    remove_wait_queue(&udsl_wqh3, &wait);
    usb_free_urb(urb);
    }
  else {
    if (!interruptible_sleep_on_timeout(&udsl_wqh3, UDSL_SEND_BULK_TIMEOUT * HZ / 1000)) {
      PDEBUG("udsl_usb_send_bulk timeout\n");
      err = -ETIME;
      }
    else
      if (signal_pending(current)) {
        PDEBUG("udsl_usb_send_bulk signal pending\n");
        err = -EINTR;
        }
    if (urb->status == -EINPROGRESS)
      usb_unlink_urb(urb);
    usb_free_urb(urb);
    remove_wait_queue(&udsl_wqh3, &wait);
    }
  return err;
  }

void udsl_usb_read_bulk_complete(urb_t *urb) {
  int *len = (int *)urb->context;
  
  //PDEBUG("udsl_usb_read_bulk_complete called\n");

  switch (urb->status) {
    case 0:
      *len = urb->actual_length;
      break;
    case -EPIPE:
      usb_clear_halt(urb->dev, usb_rcvbulkpipe(urb->dev, UDSL_ENDPOINT_INFO_IN));
      *len = 0;
      break;
    default:
      *len = 0; 
      break;
    }

  wake_up(&udsl_wqh3);
  }

int udsl_usb_read_bulk(struct udsl_instance_data *instance, char *buf, int *len) {
  urb_t *urb;
  int err;
  DECLARE_WAITQUEUE(wait, current);

  //PDEBUG("udsl_usb_read_bulk called\n");

  memset(buf, 0, *len);

  urb = usb_alloc_urb(0);
  if (!urb) { 
    len = 0;
    return -ENOMEM;
    }

  add_wait_queue(&udsl_wqh3, &wait);

  FILL_BULK_URB(urb,
   instance->usb_dev,
   usb_rcvbulkpipe(instance->usb_dev, UDSL_ENDPOINT_INFO_IN),
   buf,
   *len,
   (usb_complete_t)udsl_usb_read_bulk_complete,
   len
   );

  urb->transfer_flags |= USB_QUEUE_BULK;

  err = usb_submit_urb(urb);

  if (err) {
    remove_wait_queue(&udsl_wqh3, &wait);
    usb_free_urb(urb);
    }
  else {
    if (!interruptible_sleep_on_timeout(&udsl_wqh3, UDSL_READ_BULK_TIMEOUT * HZ / 1000)) {
      PDEBUG("udsl_usb_read_bulk timeout\n");
      err = -ETIME;
      }
    else
      if (signal_pending(current)) {
        PDEBUG("udsl_usb_read_bulk signal pending\n");
        err = -EINTR;
        }
    if (urb->status == -EINPROGRESS)
      usb_unlink_urb(urb);
    usb_free_urb(urb);
    remove_wait_queue(&udsl_wqh3, &wait);
    }

  if (err)
    len = 0;

  return err;
  }

int udsl_usb_write_status_byte(struct udsl_instance_data *instance, unsigned int address, char value) {
  char bufout[3];
  int i;

  //PDEBUG("udsl_usb_write_status_byte called\n");
  for (i = 0; i < UDSL_WRITE_STATUS_RETRIES; i++) {
    bufout[0] = (char) (address >> 8);
    bufout[1] = (char) address;
    bufout[2] = value;
    if (!udsl_usb_send_bulk(instance, bufout, 3))
      return 0;
    }

  return -1;
  }

int udsl_usb_read_status_byte(struct udsl_instance_data *instance, unsigned int address, char *value) {
  char bufout[2];
  char bufin[3];
  int len;
  int i, j;

  //PDEBUG("udsl_usb_read_status_byte called\n");
  for (i = 0; i < UDSL_READ_STATUS_RETRIES; i++) {
    bufout[0] = (char) (address >> 8);
    bufout[1] = (char) address;
    if (!udsl_usb_send_bulk(instance, bufout, 2))
      for (j = 0; j < UDSL_READ_STATUS_RETRIES; j++) {
        len = 3;
        if (!udsl_usb_read_bulk(instance, bufin, &len))
          if (bufin[0] == bufout[0] && bufin[1] == bufout[1] && len == 3) {
            *value = bufin[2]; 
            return 0;
          }
        }
    }

  return -1;
 }

int udsl_usb_read_status_word(struct udsl_instance_data *instance, unsigned int addressMSB, unsigned int addressLSB, short *value) {
 unsigned char bufMSB, bufLSB;

 if (udsl_usb_read_status_byte(instance, addressMSB, &bufMSB))
   return -1;
 if (udsl_usb_read_status_byte(instance, addressLSB, &bufLSB))
   return -1;
 if (value == NULL)
   return -1;
 else {
   if ((bufMSB == 0xff) || (bufLSB == 0xff))
     *value = 0xffff;
   else
     *value = ((bufMSB << 8) | bufLSB) & 0xffff;
 }

 return 0;
 }

int udsl_usb_set_led_state(struct udsl_instance_data *instance, char value) {
  char current_state_leds;

  PDEBUG("udsl_usb_set_led_state called\n");
  if (udsl_usb_read_status_byte(instance, LED_REGISTER, &current_state_leds))
    return -1;
  current_state_leds &= ~(LED_USB_ON | LED_ADSL_ON | LED_DATA_ON);
  current_state_leds |= value;
  if (udsl_usb_write_status_byte(instance, LED_REGISTER, current_state_leds))
    return -1;

  return 0;
  }

int udsl_usb_wait_eeprom(struct udsl_instance_data *instance, unsigned char command) {
  int i = 0;
  unsigned char status;

  //PDEBUG("udsl_usb_wait_eeprom called\n");
  do {
    i++;
    if (udsl_usb_read_status_byte(instance, I2CS, &status))
      return -1;
  } while ((!(status & command)) && (i < 4));
  if (i != 4)
    return 0;
  else
    return -1;
  }

int udsl_usb_read_eeprom(struct udsl_instance_data *instance, char mac[6]) {
  long i;
  unsigned char status;

  //PDEBUG("udsl_usb_read_eeprom called\n");
  if (udsl_usb_write_status_byte(instance, I2CS, I2CS_START))
    return -1;
  if (udsl_usb_write_status_byte(instance, I2DAT, I2CS_WRITE_ADDRESS))
    return -1;
  if (udsl_usb_wait_eeprom(instance, I2CS_ACK))
    return -1;
  if (udsl_usb_write_status_byte(instance, I2DAT, 0x10))
    return -1;
  if (udsl_usb_wait_eeprom(instance, I2CS_ACK))
    return -1;
  if (udsl_usb_write_status_byte(instance, I2CS, I2CS_STOP))
    return -1;
  if (udsl_usb_write_status_byte(instance, I2CS, I2CS_START))
    return -1;
  if (udsl_usb_write_status_byte(instance, I2DAT, I2CS_READ_ADDRESS))
    return -1;
  if (udsl_usb_wait_eeprom(instance, I2CS_ACK))
    return -1;

  if (udsl_usb_read_status_byte(instance, I2DAT, &status))
    return -1;

  for (i = 16; i <= 21; i++) {
    if (udsl_usb_read_status_byte(instance, I2DAT, &status))
      return -1;
    mac[i-16] = status;
    instance->modem_info.mac[i-16] = status;
    }
  for (i = 22; i <= 31; i++) {
    if (udsl_usb_read_status_byte(instance, I2DAT, &status))
      return -1;
    }
  for (i = 32; i <= 43; i++) {
    if (udsl_usb_read_status_byte(instance, I2DAT, &status))
      return -1;
    instance->modem_info.serial[i-32] = status;
    }
  for (i = 44; i <= 47; i++) {
    if (udsl_usb_read_status_byte(instance, I2DAT, &status))
      return -1;
    }
  for (i = 48; i <= 48; i++) {
    if (udsl_usb_read_status_byte(instance, I2DAT, &status))
      return -1;
    if (status != 0xff)
      instance->modem_info.hardversion = status;
    }
  for (i = 49; i <= 63; i++) {
    if (udsl_usb_read_status_byte(instance, I2DAT, &status))
      return -1;
    }
  for (i = 64; i <= 67; i++) {
    if (udsl_usb_read_status_byte(instance, I2DAT, &status))
      return -1;
    if (status != 0xff)
      instance->modem_info.harddata[i-64] = status;
    }
  if (udsl_usb_read_status_byte(instance, Firmware_Major_Version, &status))
    return -1;
  instance->modem_info.firmmajor = status;
  if (udsl_usb_read_status_byte(instance, Firmware_Minor_Version, &status))
    return -1;
  instance->modem_info.firmminor = status;

  if (udsl_usb_write_status_byte(instance, I2CS, I2CS_LASTREAD))
    return -1;
  if (udsl_usb_wait_eeprom(instance, I2CS_ACK))
    return -1;
  if (udsl_usb_read_status_byte(instance, I2CS, &status))
    return -1;
  if (udsl_usb_wait_eeprom(instance, I2CS_DONE))
    return -1;
  if (udsl_usb_read_status_byte(instance, I2CS, &status))
    return -1;

  return 0;
  }

int udsl_usb_read_bitrate(struct udsl_instance_data *instance) {
  unsigned short statusWord;

  instance->modem_info.downbitrate = 0;
  instance->modem_info.upbitrate = 0;
  instance->modem_info.datapath = 0;

  if (udsl_usb_read_status_word(instance, R_ChanDataIntNear_MSB, R_ChanDataIntNear_LSB, &statusWord))
    return -1;
  if ((statusWord != 0) && (statusWord != 0xffff)) {
    instance->modem_info.downbitrate = statusWord * 1024;
    instance->modem_info.datapath = InterleavedPath;
    }

  if (udsl_usb_read_status_word(instance, R_ChanDataFastNear_MSB, R_ChanDataFastNear_LSB, &statusWord))
    return -1;
  if ((statusWord != 0) && (statusWord != 0xffff)) {
    instance->modem_info.downbitrate = statusWord * 1024;
    instance->modem_info.datapath = FastPath;
    }

  if (udsl_usb_read_status_word(instance, R_ChanDataIntFar_MSB, R_ChanDataIntFar_LSB, &statusWord))
    return -1;
  if ((statusWord != 0) && (statusWord != 0xffff)) {
    instance->modem_info.upbitrate = statusWord * 1024;
    instance->modem_info.datapath = InterleavedPath;
    }

  if (udsl_usb_read_status_word(instance, R_ChanDataFastFar_MSB, R_ChanDataFastFar_LSB, &statusWord))
    return -1;
  if ((statusWord != 0) && (statusWord != 0xffff)) {
    instance->modem_info.upbitrate = statusWord * 1024;
    instance->modem_info.datapath = FastPath;
    }

  return 0;
  }

int udsl_usb_read_additional_info(struct udsl_instance_data *instance) {
  unsigned short statusWord;
  unsigned char status;

  if (instance->modem_info.datapath == InterleavedPath) {
    if (udsl_usb_read_status_word(instance, R_near_end_fec_i_2, R_near_end_fec_i_1, &statusWord))
      return -1; 
    if (statusWord && statusWord != 0xffff)
      instance->modem_info.downcorrectederrors = statusWord;

    if (udsl_usb_read_status_word(instance, R_near_end_crc_i_2, R_near_end_crc_i_1, &statusWord))
      return -1; 
    if (statusWord && statusWord != 0xffff)
      instance->modem_info.downuncorrectederrors = statusWord;

    if (udsl_usb_read_status_word(instance, R_near_end_hec_i_2, R_near_end_hec_i_1, &statusWord))
      return -1; 
    if (statusWord && statusWord != 0xffff)
      instance->modem_info.downhecerrors = statusWord;

    if (udsl_usb_read_status_word(instance, R_far_end_fec_i_2, R_far_end_fec_i_1, &statusWord))
      return -1; 
    if (statusWord && statusWord != 0xffff)
      instance->modem_info.upcorrectederrors = statusWord;

    if (udsl_usb_read_status_word(instance, R_far_end_crc_i_2, R_far_end_crc_i_1, &statusWord))
      return -1; 
    if (statusWord && statusWord != 0xffff)
      instance->modem_info.upuncorrectederrors = statusWord;

    if (udsl_usb_read_status_word(instance, R_far_end_hec_i_2, R_far_end_hec_i_1, &statusWord))
      return -1; 
    if (statusWord && statusWord != 0xffff)
      instance->modem_info.uphecerrors = statusWord;
    }
  else
    if (instance->modem_info.datapath == FastPath) {
      if (udsl_usb_read_status_word(instance, R_near_end_fec_ni_2, R_near_end_fec_ni_1, &statusWord))
        return -1; 
      if (statusWord && statusWord != 0xffff)
        instance->modem_info.downcorrectederrors = statusWord;

      if (udsl_usb_read_status_word(instance, R_near_end_crc_ni_2, R_near_end_crc_ni_1, &statusWord))
        return -1; 
      if (statusWord && statusWord != 0xffff)
        instance->modem_info.downuncorrectederrors = statusWord;

      if (udsl_usb_read_status_word(instance, R_near_end_hec_ni_2, R_near_end_hec_ni_1, &statusWord))
        return -1; 
      if (statusWord && statusWord != 0xffff)
        instance->modem_info.downhecerrors = statusWord;

      if (udsl_usb_read_status_word(instance, R_far_end_fec_ni_2, R_far_end_fec_ni_1, &statusWord))
        return -1; 
      if (statusWord && statusWord != 0xffff)
        instance->modem_info.upcorrectederrors = statusWord;

      if (udsl_usb_read_status_word(instance, R_far_end_crc_ni_2, R_far_end_crc_ni_1, &statusWord))
        return -1; 
      if (statusWord && statusWord != 0xffff)
        instance->modem_info.upuncorrectederrors = statusWord;

      if (udsl_usb_read_status_word(instance, R_far_end_hec_ni_2, R_far_end_hec_ni_1, &statusWord))
        return -1; 
      if (statusWord && statusWord != 0xffff)
        instance->modem_info.uphecerrors = statusWord;
      }

  if (udsl_usb_read_status_byte(instance, R_attenuationDnStr, &status))
    return -1;
  if (status != 0xff)
    instance->modem_info.downattenuation = status/2;
  else
    instance->modem_info.downattenuation = 0;

  if (udsl_usb_read_status_byte(instance, R_attenuationUpStr, &status))
    return -1;
  if (status != 0xff)
    instance->modem_info.upattenuation = status/2;
  else
    instance->modem_info.upattenuation = 0;

  if (udsl_usb_read_status_byte(instance, R_noiseMarginDnstr, &status))
    return -1;
  if (status != 0xff)
    instance->modem_info.downsignalnoiseratio = status/2;
  else
    instance->modem_info.downsignalnoiseratio = 0;

  if (udsl_usb_read_status_byte(instance, R_noiseMarginUpstr, &status))
    return -1;
  if (status != 0xff)
    instance->modem_info.upsignalnoiseratio = status/2;
  else
    instance->modem_info.upsignalnoiseratio = 0;

  if (udsl_usb_read_status_byte(instance, R_OutputPowerDnstr, &status))
    return -1;
  if (status != 0xff)
    instance->modem_info.downoutputpower = status/2;
  else
    instance->modem_info.downoutputpower = 0;

  if (udsl_usb_read_status_byte(instance, R_OutputPowerUpstr, &status))
    return -1;
  if (status != 0xff)
    instance->modem_info.upoutputpower = status/2;
  else
    instance->modem_info.upoutputpower = 0;

  return 0;
  }

int udsl_usb_connect_line (struct udsl_instance_data *instance) {
  if (udsl_usb_write_status_byte(instance, 0x1cbd, 0x00))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x81fd, 0x0d))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x804a, 0x01))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x804b, 0x00))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x804c, 0x0f))
    return -1;
  if (open != ADSL_Open && open != ADSL_Open_ANSI && open != ADSL_Open_GDMT && open != ADSL_Open_GLITE && open != ADSL_Open_MULTIMODE)
    open = ADSL_Open_ANSI;
  if (udsl_usb_write_status_byte(instance, TX_COMMAND, open))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x1cbd, 0x00))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x1cbc, 0x00))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x7fae, 0x1f))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x7fab, 0x02))
    return -1;

  return 0;
  }

int udsl_usb_close_line (struct udsl_instance_data *instance) {
  
  udsl_usb_set_led_state(instance, LED_USB_ON);
  if (udsl_usb_write_status_byte(instance, TX_COMMAND, ADSL_Close))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x1cbd, 0x00))
    return -1;
  if (udsl_usb_write_status_byte(instance, 0x1cbc, 0x00))
    return -1;

  return 0;
  }

void udsl_usb_monitorize (struct udsl_instance_data *instance) {
  unsigned char rx_cmd;
  unsigned char adsl_status;
  unsigned char status;

  //PDEBUG("Getting RX_COMMAND...\n");
  if (udsl_usb_read_status_byte(instance, RX_COMMAND, &rx_cmd))
    return;
  //PDEBUG("Command received = 0x%02x\n", rx_cmd);
  //PDEBUG("Getting MODEM_STATUS...\n");
  if (udsl_usb_read_status_byte(instance, MODEM_STATUS, &adsl_status))
    return;
  //PDEBUG("Status = 0x%02x\n", adsl_status);
  if (instance->modem_info.lastcmd != rx_cmd || instance->modem_info.laststatus != adsl_status) {
    switch (adsl_status) {
      case MODEM_DOWN:
        if (instance->modem_info.laststatus == MODEM_UP && rx_cmd == ADSL_Connection_Closed) {
          PDEBUG("Connection closed\n");
	  instance->modem_info.linestatus = ADSL_Connection_Closed;
	  instance->modem_info.operationalmode = 0;
          instance->modem_info.downbitrate = 0;
          instance->modem_info.upbitrate = 0;
	  instance->modem_info.datapath = 0;
          instance->adslled = OFF;
          udsl_usb_set_led_state(instance, LED_USB_ON);
          if (udsl_usb_connect_line(instance))
            return;
          }
        break;
      case MODEM_UP:
        if (rx_cmd == ADSL_Connection_Opened) {
          PDEBUG("ADSL line is up\n");
	  instance->modem_info.linestatus = ADSL_Connection_Opened;
          udsl_usb_write_status_byte(instance, TX_COMMAND, 0x31);  /* update ops mode */
          udsl_usb_read_bitrate(instance);
          instance->adslled = ON;
          udsl_usb_set_led_state(instance, LED_USB_ON | LED_ADSL_ON);
          instance->modem_info.operationalmode = 0xff;
          if (!udsl_usb_read_status_byte(instance, R_Operational_Mode, &status))
            instance->modem_info.operationalmode = status;
          }
        else if (instance->modem_info.laststatus == MODEM_UP && rx_cmd == ADSL_Connection_Lost) {
          PDEBUG("Connection lost\n");
	  instance->modem_info.linklost++;
	  instance->modem_info.linestatus = ADSL_Connection_Lost;
	  instance->modem_info.operationalmode = 0;
          instance->modem_info.downbitrate = 0;
          instance->modem_info.upbitrate = 0;
	  instance->modem_info.datapath = 0;
          instance->adslled = OFF;
          udsl_usb_set_led_state(instance, LED_USB_ON);
          if (udsl_usb_connect_line(instance))
            return;
          }
        break;
      case MODEM_INIT:
        instance->modem_info.linestatus = ADSL_In_Training;
	break;
      }
    instance->modem_info.lastcmd = rx_cmd;
    instance->modem_info.laststatus = adsl_status;
    }
    if (adsl_status == MODEM_INIT || adsl_status == MODEM_WAIT) {
      if (instance->adslled == OFF) {
        instance->adslled = ON;
        udsl_usb_set_led_state(instance, LED_USB_ON | LED_ADSL_ON);
        }
      else {
        instance->adslled = OFF;
        udsl_usb_set_led_state(instance, LED_USB_ON);
        }
      }
  if (instance->modem_info.extendedinfo == 1 && instance->modem_info.linestatus == ADSL_Connection_Opened)
    udsl_usb_read_additional_info(instance); 
  }

int udsl_usb_monitorize_thread (void *data) {
  int i=0;
  int j=UDSL_POLL_MODEM_STATUS-1;
  DECLARE_WAITQUEUE(wait, current);
     
  lock_kernel();
  exit_files(current);  /* daemonize doesn't do exit_files */
  daemonize();

  /* Setup a nice name */
  strcpy(current->comm, "kMonADSLd");
  
  add_wait_queue(&udsl_wqh2, &wait);

  for (i=0; i < MAX_UDSL ; i++)
    if (minor_data[i]) {
      minor_data[i]->modem_info.lastcmd = 0xff;
      minor_data[i]->modem_info.laststatus = MODEM_UP;
      minor_data[i]->dataled = OFF;
      minor_data[i]->adslled = ON;
      }
  
  for(;;) {
    interruptible_sleep_on_timeout(&udsl_wqh2, UDSL_POLL_MONITOR_DELAY * HZ / 1000);
    if (signal_pending(current)) break;
    //PDEBUG("kMonADSLd awoke\n");
    j++;
    for (i=0; i < MAX_UDSL ; i++)
      if (minor_data[i]) {
        if (j == UDSL_POLL_MODEM_STATUS)
          udsl_usb_monitorize(minor_data[i]);
        if (minor_data[i]->pending_data_led > 0) {
          minor_data[i]->pending_data_led--;
          if (minor_data[i]->dataled == OFF) {
            if (minor_data[i]->adslled == ON) {
              if (!udsl_usb_set_led_state(minor_data[i], LED_USB_ON | LED_ADSL_ON | LED_DATA_ON))
                minor_data[i]->dataled = ON;
              }
            else
              if (!udsl_usb_set_led_state(minor_data[i], LED_USB_ON | LED_DATA_ON))
                minor_data[i]->dataled = ON;
            }
          }
        else {
          if (minor_data[i]->dataled == ON) {
            if (minor_data[i]->adslled == ON) {
              if (!udsl_usb_set_led_state(minor_data[i], LED_USB_ON | LED_ADSL_ON))
                minor_data[i]->dataled = OFF;
              }
            else
              if (!udsl_usb_set_led_state(minor_data[i], LED_USB_ON))
                minor_data[i]->dataled = OFF;
            }
          }
        }
      if (j == UDSL_POLL_MODEM_STATUS)
        j = 0;
    }
  
  datapid2 = 0;
  PDEBUG("kMonADSLd is exiting\n");
  return 0;
  }

void udsl_usb_mon_start (void) {
  datapid2 = kernel_thread(udsl_usb_monitorize_thread, (void *)NULL,
      CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
  }

void udsl_usb_mon_stop (void) {
  int ret;

  /* Kill the thread */
  ret = kill_proc(datapid2, SIGTERM, 1);
  if (!ret) {
    /* Wait 10 seconds */
    int count = 10 * 100;

    while (datapid2 && --count) {
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(1);
      }

    if (!count)
      err("giving up on killing MonADSL thread.");
    }
  }


/***************************************************************************
*
* usb driver entries
*
****************************************************************************/
#define hex2int(c) ( (c >= '0')&&(c <= '9') ?  (c - '0') : ((c & 0xf)+9) )


static int udsl_usb_ioctl (struct usb_device *dev, unsigned int code, void *user_data) {
  struct udsl_instance_data *instance;
  int i;
  struct udsl_modem_info *modem_info;

  PDEBUG("IOCTL called\n");

  modem_info = (struct udsl_modem_info *) user_data;

  for (i=0; i < MAX_UDSL; i++)
    if (minor_data[i] && (minor_data[i]->usb_dev == dev))
      break;
  
  if (i==MAX_UDSL)
    return -EINVAL;

  instance = minor_data[i];

  switch(code) {
    case UDSL_IOCTL_START:
      return udsl_usb_data_init (instance);
      break;
    case UDSL_IOCTL_STOP:
      return udsl_usb_data_exit (instance);
      break;
    case UDSL_IOCTL_INFO:
      memcpy(modem_info, &instance->modem_info, sizeof(struct udsl_modem_info));
      return 0;
      break;
    case UDSL_IOCTL_RECONNECT:
      if (speed == MAX_SPEED)
        return udsl_usb_connect_line(instance);
      else
        return -EINVAL;
      break;
    case UDSL_IOCTL_UPDATE_OPS_MODE:
      if (speed == MAX_SPEED) {
        if (udsl_usb_write_status_byte(instance, TX_COMMAND, 0x31))
	  return -1;
        else
	  return 0;
        }
      else
	return -EINVAL;
      break;
    case UDSL_IOCTL_START_EXTENDED_INFO:
      if (speed == NORMAL_SPEED) {
        instance->modem_info.extendedinfo = 1;
        return 0;
      }
      else
	return -1;
      break;
    case UDSL_IOCTL_STOP_EXTENDED_INFO:
      if (speed == NORMAL_SPEED) {
        instance->modem_info.extendedinfo = 0;
        return 0;
      }
      else
	return -1;
      break;
    case UDSL_IOCTL_CLEAR_COUNTERS:
      instance->modem_info.atmcelltx = 0;
      instance->modem_info.atmcellrx = 0;
      instance->modem_info.oamloopback = 0;
      instance->modem_info.linklost = 0;
      return 0;
      break;
    default:
      break;
    }
  return -EINVAL;
  }

void *udsl_usb_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id)
{
  int i;
  unsigned char mac[6];
  struct udsl_instance_data *instance = NULL;
  unsigned char status;

  PDEBUG("Trying device with Vendor=0x%x, Product=0x%x, ifnum %d\n", dev->descriptor.idVendor, dev->descriptor.idProduct, ifnum);

  if  ((dev->descriptor.bDeviceClass != USB_CLASS_VENDOR_SPEC)||
      (dev->descriptor.idVendor  != CP4218_VENDORID)||
      (dev->descriptor.idProduct != CP4218_PRODUCTID)||
      (ifnum!=0))
    return NULL;

  MOD_INC_USE_COUNT;

  for (i=0; i < MAX_UDSL; i++)
    if (minor_data[i] == NULL)
      break;

  if (i >= MAX_UDSL) {
    printk(KERN_INFO "No minor table space available for 3Com ADSL Modem USB\n");
    MOD_DEC_USE_COUNT;
    return NULL;
    };

  PDEBUG("Device Accepted, assigning minor %d\n", i);

  /* device init */
  instance = kmalloc(sizeof(struct udsl_instance_data), GFP_ATOMIC);
  if (!instance) {
    PDEBUG("No memory for Instance data!\n");
    MOD_DEC_USE_COUNT;
    return NULL;
    }

  /* initialize structure */
  memset(instance, 0, sizeof(struct udsl_instance_data));  
  instance->minor    = i;
  instance->usb_dev  = dev;
  instance->rcvbufs  = NULL;
  instance->modem_info.modminor = MOD_MINOR_VERSION;
  instance->modem_info.modmajor = MOD_MAJOR_VERSION;
  instance->modem_info.modspeed = speed;
  instance->modem_info.openmode = open;
  if (udsl_usb_read_status_byte(instance, DSP_Booted, &status))
    if (status != 1) {
      PDEBUG("Error: You need load Alcatel firmware\n");
      kfree(instance);
      MOD_DEC_USE_COUNT;
      return NULL;
      }
  if (udsl_usb_read_eeprom(instance, mac)) {
    PDEBUG("I can't read serial number\n");
    kfree(instance);
    MOD_DEC_USE_COUNT;
    return NULL;
    }
  if (udsl_usb_read_status_byte(instance, SW_Version_LSB, &status)) {
    PDEBUG("Error reading Alcatel firmware minor version\n");
    kfree(instance);
    MOD_DEC_USE_COUNT;
    return NULL;
    }
  instance->modem_info.alcfirmminor = status;
  if (udsl_usb_read_status_byte(instance, SW_Version_MSB, &status)) {
    PDEBUG("Error reading Alcatel firmware major version\n");
    kfree(instance);
    MOD_DEC_USE_COUNT;
    return NULL;
    }
  instance->modem_info.alcfirmmajor = status;
  instance->modem_info.operationalmode = 0xff;
  if (udsl_usb_read_status_byte(instance, MODEM_STATUS, &status)) {
    PDEBUG("Error reading modem status\n");
    kfree(instance);
    MOD_DEC_USE_COUNT;
    return NULL;
    }
  instance -> modem_info.laststatus = status;
  if (udsl_usb_read_status_byte(instance, RX_COMMAND, &status)) { 
    PDEBUG("Error reading received command\n");
    kfree(instance);
    MOD_DEC_USE_COUNT;
    return NULL;
    }
  instance->modem_info.lastcmd = status;
  if (instance->modem_info.laststatus == MODEM_UP && (instance->modem_info.lastcmd == ADSL_Connection_Opened || instance->modem_info.lastcmd == ADSL_Operational_Mode_Ack)) {
    instance->modem_info.linestatus = ADSL_Connection_Opened;
    if (udsl_usb_set_led_state(instance, LED_USB_ON | LED_ADSL_ON)) {
      PDEBUG("Error setting ADSL led on\n");
      kfree(instance);
      MOD_DEC_USE_COUNT;
      return NULL;
      }
    if (udsl_usb_read_status_byte(instance, R_Operational_Mode, &status)) {
      PDEBUG("Error reading operational mode\n");
      kfree(instance);
      MOD_DEC_USE_COUNT;
      return NULL;
      }
    instance->modem_info.operationalmode = status;
    if (udsl_usb_read_bitrate(instance)) {
      PDEBUG("Error reading bitrate\n");
      kfree(instance);
      MOD_DEC_USE_COUNT;
      return NULL;
      }
    }
  else {
    if (instance->modem_info.laststatus == MODEM_INIT)
      instance->modem_info.linestatus = ADSL_In_Training;
    else
      instance->modem_info.linestatus = ADSL_Connection_Closed;
    if (udsl_usb_set_led_state(instance, LED_USB_ON)) {
      PDEBUG("Error setting ADSL led off\n");
      kfree(instance);
      MOD_DEC_USE_COUNT;
      return NULL;
      }
    }
    
  spin_lock_init (&instance->sndqlock);
  spin_lock_init (&instance->recvqlock);

  udsl_atm_startdevice(instance, &udsl_atm_devops);

  PDEBUG("MAC is %02x:%02x:%02x:%02x:%02x:%02x\n", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
  udsl_atm_set_mac(instance, mac);

  minor_data[instance->minor] = instance;

  if (speed != MAX_SPEED) {
    instance->pending_data_led = 0;
    udsl_usb_mon_start();
    }

  return instance;
}

void udsl_usb_disconnect(struct usb_device *dev, void *ptr)
{
  struct udsl_instance_data *instance = (struct udsl_instance_data *)ptr;
  int i = instance->minor;

  if (speed != MAX_SPEED)
    udsl_usb_mon_stop();

  /* unlinking receive buffers */
  udsl_usb_data_exit(instance);

  /* close line */
  udsl_usb_close_line(instance);
  
  /* removing atm device */
  if (instance->atm_dev)
    udsl_atm_stopdevice(instance);

  PDEBUG("disconnecting minor %d\n", i);

  while (MOD_IN_USE > 1) {
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(1);
    }

  kfree (instance);
  minor_data[i] = NULL;

  MOD_DEC_USE_COUNT;
}

/***************************************************************************
*
* Driver Init
*
****************************************************************************/

int udsl_usb_init(void)
{
  int i;

  PDEBUG ("Initializing 3Com ADSL Modem USB Driver Version %d.%d\n", MOD_MAJOR_VERSION, MOD_MINOR_VERSION);
  
  for (i =0; i < MAX_UDSL ; i++)
    minor_data[i] = NULL;

  init_waitqueue_head(&udsl_wqh);
  init_waitqueue_head(&udsl_wqh2);
  init_waitqueue_head(&udsl_wqh3);
  init_waitqueue_head(&udsl_wqh4);
  udsl_atm_sar_start();
  
  return usb_register(&udsl_usb_driver);
}

int udsl_usb_cleanup(void)
{
  /* killing threads */
  udsl_atm_sar_stop();
  
  usb_deregister(&udsl_usb_driver);
  return 0;
}

#ifdef MODULE
int init_module(void)
{
	return udsl_usb_init();
}

int cleanup_module(void)
{ 
	return udsl_usb_cleanup();
}
#endif

#ifdef DEBUG_PACKET
/*******************************************************************************
*
* Debug 
*
*******************************************************************************/

int udsl_print_packet (const unsigned char *data, int len) {
  unsigned char buffer[2048];
  int i = 0, j = 0;
  int pos;
  int nlin; 

  if (!data) {
    PDEBUG("udsl_print_packet NULL pointer\n");
    return 0;
    }
  if (len < 0) {
    PDEBUG("udsl_print_packet invalid length\n");
    return 0;
    }
  nlin = (len / 16) + 1;
  if ((nlin * 71) > 1024) {  /* 71 = 6 + 1 + 16 * 3 + 16 */
    PDEBUG("udsl_print_packet length exceeds buffer\n");
    return 0;
    }

  for (i = 0; i < len ; ) {
    buffer[0] = '\0';
    sprintf(buffer,"%.3d :", i);
    pos = i;
    for (j = 0; (j < 16) && (i < len) ; j++, i++) 
      sprintf(buffer, "%s %2.2x", buffer, data[i]);
    for (; j < 16; j++) 
      sprintf(buffer, "%s   ", buffer);
    sprintf(buffer, "%s ", buffer);
    for (j = 0; (j < 16) && (pos < len) ; j++, pos++) {
      if (data[pos] >= ' ' && data[pos] <= 0x7f)
        sprintf(buffer, "%s%c", buffer, data[pos]);
      else
        sprintf(buffer, "%s.", buffer);
      }
    PDEBUG("%s\n", buffer);
    }
  return i;
  }

#endif /* PACKETDEBUG */

unsigned char *pad_data(unsigned char *source, int l) {
  int number_of_cells;
  int total_length;
  unsigned char *out;
  unsigned char *target;
  int i;

  if (!source || l < 53 || (l % 53 != 0))
    return NULL;
  number_of_cells = l / 53;
  total_length = number_of_cells * 64;
  out = kmalloc (total_length, GFP_ATOMIC);
  if (!out)
    return NULL;
  target = out;
  for (i = 0 ; i < number_of_cells; i++) {
    memcpy(target, source, 53);
    target += 53;
    memset(target, 0, 11);
    target += 11;
    source += 53;
    }
  
  return out; 
  }

/* CRC-10 table */
unsigned short byte_crc10_table[256] = {
  0x000, 0x233, 0x255, 0x066, 0x299, 0x0aa, 0x0cc, 0x2ff,
  0x301, 0x132, 0x154, 0x367, 0x198, 0x3ab, 0x3cd, 0x1fe,
  0x031, 0x202, 0x264, 0x057, 0x2a8, 0x09b, 0x0fd, 0x2ce,
  0x330, 0x103, 0x165, 0x356, 0x1a9, 0x39a, 0x3fc, 0x1cf,
  0x062, 0x251, 0x237, 0x004, 0x2fb, 0x0c8, 0x0ae, 0x29d,
  0x363, 0x150, 0x136, 0x305, 0x1fa, 0x3c9, 0x3af, 0x19c,
  0x053, 0x260, 0x206, 0x035, 0x2ca, 0x0f9, 0x09f, 0x2ac,
  0x352, 0x161, 0x107, 0x334, 0x1cb, 0x3f8, 0x39e, 0x1ad,
  0x0c4, 0x2f7, 0x291, 0x0a2, 0x25d, 0x06e, 0x008, 0x23b,
  0x3c5, 0x1f6, 0x190, 0x3a3, 0x15c, 0x36f, 0x309, 0x13a,
  0x0f5, 0x2c6, 0x2a0, 0x093, 0x26c, 0x05f, 0x039, 0x20a,
  0x3f4, 0x1c7, 0x1a1, 0x392, 0x16d, 0x35e, 0x338, 0x10b,
  0x0a6, 0x295, 0x2f3, 0x0c0, 0x23f, 0x00c, 0x06a, 0x259,
  0x3a7, 0x194, 0x1f2, 0x3c1, 0x13e, 0x30d, 0x36b, 0x158,
  0x097, 0x2a4, 0x2c2, 0x0f1, 0x20e, 0x03d, 0x05b, 0x268,
  0x396, 0x1a5, 0x1c3, 0x3f0, 0x10f, 0x33c, 0x35a, 0x169,
  0x188, 0x3bb, 0x3dd, 0x1ee, 0x311, 0x122, 0x144, 0x377,
  0x289, 0x0ba, 0x0dc, 0x2ef, 0x010, 0x223, 0x245, 0x076,
  0x1b9, 0x38a, 0x3ec, 0x1df, 0x320, 0x113, 0x175, 0x346,
  0x2b8, 0x08b, 0x0ed, 0x2de, 0x021, 0x212, 0x274, 0x047,
  0x1ea, 0x3d9, 0x3bf, 0x18c, 0x373, 0x140, 0x126, 0x315,
  0x2eb, 0x0d8, 0x0be, 0x28d, 0x072, 0x241, 0x227, 0x014,
  0x1db, 0x3e8, 0x38e, 0x1bd, 0x342, 0x171, 0x117, 0x324,
  0x2da, 0x0e9, 0x08f, 0x2bc, 0x043, 0x270, 0x216, 0x025,
  0x14c, 0x37f, 0x319, 0x12a, 0x3d5, 0x1e6, 0x180, 0x3b3,
  0x24d, 0x07e, 0x018, 0x22b, 0x0d4, 0x2e7, 0x281, 0x0b2,
  0x17d, 0x34e, 0x328, 0x11b, 0x3e4, 0x1d7, 0x1b1, 0x382,
  0x27c, 0x04f, 0x029, 0x21a, 0x0e5, 0x2d6, 0x2b0, 0x083,
  0x12e, 0x31d, 0x37b, 0x148, 0x3b7, 0x184, 0x1e2, 0x3d1,
  0x22f, 0x01c, 0x07a, 0x249, 0x0b6, 0x285, 0x2e3, 0x0d0,
  0x11f, 0x32c, 0x34a, 0x179, 0x386, 0x1b5, 0x1d3, 0x3e0,
  0x21e, 0x02d, 0x04b, 0x278, 0x087, 0x2b4, 0x2d2, 0x0e1};

unsigned short update_crc10_by_bytes(unsigned short crc10_accum, unsigned char *data_blk_ptr, int  data_blk_size) {
  register int i;
  for ( i = 0;  i < data_blk_size;  i++ ) {
    crc10_accum = ((crc10_accum << 8) & 0x3ff)
      ^ byte_crc10_table[( crc10_accum >> 2) & 0xff]
      ^ *data_blk_ptr++;
    }

  return crc10_accum;
  }

void send_complete(urb_t *urb) {
  
  PDEBUG("send_complete called\n");
  usb_free_urb(urb);
  }

int remove_control_cells(unsigned char *source, int l, unsigned char *out, struct udsl_instance_data *instance) {
  int number_of_cells;
  int total_length;
  unsigned char *target;
  int i;
  int control_cells = 0;
  short pti;
  short vpi;
  int vci;
  unsigned short crc10_accum;
  urb_t *urb;
  int err;
  unsigned char *buf;

  if (!source || l < 53 || (l % 53 != 0) || !out)
    return -1;

  PDEBUG("Before remove control cells:\n");
  PACKETDEBUG(source, l);

  target = out;
  buf = instance->oambuff;
  number_of_cells = l / 53;
  for (i = 0 ; i < number_of_cells; i++) {
    pti = (source[(i*53)+3] >> 1) & 0x07;
    if (pti & ATM_PT_TYPE) {  /* is not user cell? */
      control_cells++;
      if ((pti == ATM_PTI_E2EF5) && (source[(i*53)+5] == OAM_LOOPBACK)) {  /* is OAM LOOPBACK cell? */
	if (source[(i*53)+6] == 1) { /* return OAM LOOPBACK cell? */
	  vpi = ((source[i*53] << 4) | (source[(i*53)+1] >> 4)) & 0xff;
	  vci = ((source[(i*53)+1] << 12) | (source[(i*53)+2] << 4) | (source[(i*53)+3] >> 4)) & 0xffff;
	  if ((instance->modem_info.vpi == vpi) && (instance->modem_info.vci == vci)) {  /* is the cell directed to our VPI.VCI? */
  	    // TODO: check vpi, vci in vcc list
            PDEBUG("OAM LOOPBACK cell detected:\n");
            PACKETDEBUG(source+(i*53), 53);
	    /* check CRC-10 */
	    memset(buf, 0, 64);
	    memcpy(buf, source+(i*53), 53);
            crc10_accum = update_crc10_by_bytes(0, (unsigned char *) &buf[5], 48);
	    if (crc10_accum != 0)
	      PDEBUG("Corruption detected in OAM LOOPBACK cell\n");
	    else {
	      instance->modem_info.oamloopback++;
              urb = usb_alloc_urb(0);
  	      if (urb) {
  	        buf[6]--;  /* unset loopback */
       	        buf[51] = 0;  /* if set last two 2 bytes (CRC) to zero then crc still is valid */
	        buf[52] = 0;
                crc10_accum = update_crc10_by_bytes(0, (unsigned char *) &buf[5], 48);	  
	        buf[51] ^= crc10_accum >> 8;
	        buf[52] ^= crc10_accum & 0xff;
                FILL_BULK_URB(urb,
                 instance->usb_dev,
                 usb_sndbulkpipe(instance->usb_dev, UDSL_ENDPOINT_DATA_OUT),
                 buf,
                 64,
                 (usb_complete_t)send_complete,
                 instance
                );
                urb->transfer_flags |= USB_QUEUE_BULK;

                PDEBUG("OAM LOOPBACK cell to return:\n");
                PACKETDEBUG(buf, 64);

                err = usb_submit_urb(urb);
                if (err)
                  usb_free_urb(urb);
	        }
	      }
	    }
	  }
        }
      }
    else {
      memcpy(target, source+(i*53), 53);
      target += 53;
      }
    }

  total_length = (number_of_cells - control_cells) * 53;
  if (total_length == 0)
    return 0;

  PDEBUG("After remove control cells:\n");
  PACKETDEBUG(out, total_length);

  return total_length;
  }
